-- PROJECT:  TYPE.EX; Learn to type (using US keyboard layout).
-- AUTHOR:   Shian Lee; http://qe.site90.com
-- VERSION:  1.00  4/13/2011
-- LANGUAGE: Euphoria programming language version 3.1.1
-- DETAILS:  To run this program:
--           * Get Euphoria version 3.1.1 or version 4.x from
--             http://www.RapidEuphoria.com;
--           * Get QE editor from http://qe.site90.com, open this source file
--             and press F5 to start.
-- DISCLAIMER: Use at your own risk (you may report problem at QE site).
-- ==========================================================================



-- all include files:
include graphics.e              -- text_color()...
include get.e                   -- wait_key()...
include wildcard.e              -- upper()...



-- all constants:
constant
    -- program title & version:
    PROGRAM_TITLE = "Learn to type v1.0",

    -- program's colors:
    BACK_COLOR = BLACK,
    TXT_COLOR = WHITE,
    BACK_NAIL_COLOR = CYAN,
    TEXT_NAIL_COLOR = BLACK,

    -- booleans:
    FALSE = 0, TRUE = 1,
    -- 1 = standard output:
    SCREEN = 1,
    -- keyboard keys:
    ENTER = 13, F1 = 315, F5 = 319, F10 = 324,
    -- lesson's string length (in the edit box):
    STRING_LENGTH = 66,
    -- left column to start print the LESSON_SCREEN constant lines:
    LESSON_COLUMN = 6,

    -- help screen:
    HELP_SCREEN = {
	PROGRAM_TITLE & " - Help:",
	repeat('', length(PROGRAM_TITLE) + 8),
	"",
	"1. Turn off the <CapsLock> key before you start.",
	"2. Choose lesson number using <F5> key.",
	"3. Put your hands on the keyboard in 'Start Position':",
	"   The fingers next to the thumbs on <F> and <J> keys (left hand for F",
	"   and right hand for J).",
	"4. You may practice a lesson as long as you like, and when you feel ready",
	"   then try the next lesson in order (the following number).",
	"5. Quit the program with <F10> key.",
	"",
	"Notes:",
	"* The lessons based on US keyboard layout.",
	"* Use the <Shift> keys for capital letters.",
	"* Press <Shift/Backspace/Enter/Tab/CapsLock> keys with the small fingers.",
	"* Press <Space> with one of your thumbs.",
	"* Press <Esc/F1..F12/Ctrl/Alt> and other special keys - any way you want.",
	"",
	"",
	"Enjoy and good luck!      http://qe.site90.com",
	"",
	"Press any key to exit help..."
    },

    -- lesson screen (shape size is 24x70):
    LESSON_SCREEN = {
	" Lesson",
	"ͻ",
	"                                                                    ",
	"ͼ",
	"       Ŀ",
	"        ~  !  @  #  $  %  ^  &  *  (  )  _  +  <--",
	"        `  1  2  3  4  5  6  7  8  9  0  -  =  BkSp",
	"       Ŀ",
	"           Q  W  E  R  T  Y  U  I  O  P  {  }  | ",
	"      Tab  q  w  e  r  t  y  u  i  o  p  [  ]  \\ ",
	"          ",
	"      Caps  A  S  D  F  G  H  J  K  L  :  \" ",
	"      Lock  a  s  d  f  g  h  j  k  l  ;  '  Enter ",
	"           ",
	"        Left   Z  X  C  V  B  N  M  <  >  ?  Right",
	"        Shift  z  x  c  v  b  n  m  ,  .  /  Shift ",
	"              ",
	"            Ŀ                Space                Ŀ",
	"          Ĵ                Ŀ",
	"            Ŀ                                   Ĵ  ",
	"        Ĵ                                         Ŀ",
	"            Ŀ                             Ŀ    ",
	"                                                   ",
	"                                                     "
    }



-- PURPOSE: set both text and back screen colors at once.
--          (just to make the code more clear).
procedure set_color(integer textColor, integer backColor)
    text_color(textColor)
    bk_color(backColor)
end procedure



-- PURPOSE: simply clear the whole screen and hide the cursor.
-- USING:   set_color
procedure clear_the_screen()
    cursor(NO_CURSOR)
    set_color(TXT_COLOR, BACK_COLOR)
    clear_screen()
end procedure



-- PURPOSE: draw the lesson screen (static shape).
-- USING:   clear_the_screen
procedure draw_lesson_screen(integer lessonNumber)
    clear_the_screen()

    -- draw the lesson screen:
    for line = 1 to length(LESSON_SCREEN) do
	position(line, LESSON_COLUMN)
	puts(SCREEN, LESSON_SCREEN[line])
    end for

    -- draw the status line:
    position(25, 3)
    puts(SCREEN, "<F1=Help>")
    position(25, 31)
    puts(SCREEN, "<F5=Choose Lesson>")
    position(25, 69)
    puts(SCREEN, "<F10=Quit>")

    -- print the current lesson number on top:
    position(1, length(LESSON_SCREEN[1]) + 7)
    puts(SCREEN, sprintf("%d:", lessonNumber))
end procedure



-- PURPOSE: draw the help screen (static shape).
-- USING:   clear_the_screen
procedure draw_help_screen()
    clear_the_screen()

    -- draw the help screen:
    for line = 1 to length(HELP_SCREEN) do
	position(line, 2)
	puts(SCREEN, HELP_SCREEN[line])
    end for

    -- show the cursor on the last line:
    position(length(HELP_SCREEN), 2 + length(HELP_SCREEN[$]))
    cursor(UNDERLINE_CURSOR)
end procedure



-- PURPOSE: let user choose a new lesson number between 1 to 18.
-- USING:   clear_the_screen
-- RETURN:  integer; lesson number between 1 to 18.
function choose_lesson()
    clear_the_screen()

    -- let user choose a new lesson number:
    position(2, 6)
    puts(SCREEN, PROGRAM_TITLE)
    position(3, 5)
    puts(SCREEN, repeat('', length(PROGRAM_TITLE) + 2))
    position(5, 5)
    cursor(UNDERLINE_CURSOR)
    return floor(prompt_number("Please enter a lesson number (1..18): ", {1, 18}))
end function



-- PURPOSE: create a random string, according to the current lesson number
--          (include keys from all lessons up to the current lesson).
-- RETURN:  sequence; {keys, fingers}  (two parallel strings)
function create_lesson_string(integer lessonNumber)
    sequence data, keys, fingers
    integer randLesson, randKey

    -- Lessons data includes parallel keys & fingers for each lesson:
    -- 10 fingers are counted from left hand to right hand,
    -- the small finger of the right hand (number 10) is using '0'.
    -- data includes only small letters - capital letters is added by code.
    -- there are 18 lessons that covers all keys (including Shift+Key).
    data = {
    --  Keys:        Finger No.:
       {"fj"       , "47"       },  -- lesson 1
       {"gh"       , "47"       },  -- lesson 2
       {"dk"       , "38"       },  -- lesson 3
       {"asl"      , "129"      },  -- lesson 4
       {"rtuy"     , "4477"     },  -- lesson 5
       {"ei"       , "38"       },  -- lesson 6
       {"qwop"     , "1290"     },  -- lesson 7
       {"cvbmn"    , "44477"    },  -- lesson 8
       {"xz"       , "32"       },  -- lesson 9
       {";'"       , "00"       },  -- lesson 10 (Start using capital letters)
       {",./"      , "890"      },  -- lesson 11
       {"[]\\"     , "000"      },  -- lesson 12
       {"5678"     , "4477"     },  -- lesson 13
       {"3490"     , "2389"     },  -- lesson 14
       {"`12-="    , "11100"    },  -- lesson 15
       {"<>?:\"{}|", "89000000" },  -- lesson 16 (Shift keys)
       {"%^&*"     , "4477"     },  -- lesson 17 (Shift keys)
       {"~!@#$()_+", "111238900"}   -- lesson 18 (Shift keys)
    }

    -- initialize the two parallel strings:
    keys = ""
    fingers = ""

    -- create one lesson string:
    while length(keys) < STRING_LENGTH do

	-- create one random "word" of length 1 to 9:
	for letter = 1 to rand(9) do
	    -- pick up random key from random lesson up to current lesson:
	    randLesson = rand(lessonNumber)
	    randKey = rand(length(data[randLesson][1]))
	    -- add key and its finger number to parallel strings:
	    keys &=    data[randLesson][1][randKey]
	    fingers &= data[randLesson][2][randKey]
	    -- from lesson 10 and up capitalize the first letter of a word:
	    if letter = 1 and lessonNumber >= 10 then
		keys[$] = upper(keys[$])
	    end if
	end for

	-- add space (ASCII 32) after the "word":
	keys &= 32
	fingers &= '6'  -- (right hand thumb - finger number 6).
    end while

    -- cut the two parallel strings, and add ENTER key at their end:
    return {keys[1..STRING_LENGTH] & ENTER, fingers[1..STRING_LENGTH] & '0'}
end function



-- PURPOSE: show the finger to use for the current key in lesson.
-- USING:   set_color
procedure draw_nail(integer key, integer finger)
    sequence nailPos

    -- nails positions data (positions within LESSON_SCREEN constant):
    nailPos = {
    --  row, column
	{22, 10},   -- finger no. 1 (Left hand small finger)
	{20, 12},   -- finger no. 2
	{19, 14},   -- finger no. 3
	{21, 16},   -- finger no. 4
	{23, 19},   -- finger no. 5 (left hand thumb)
	{23, 51},   -- finger no. 6 (Right hand thumb)
	{21, 54},   -- finger no. 7
	{19, 56},   -- finger no. 8
	{20, 58},   -- finger no. 9
	{22, 60}    -- finger no. 0 (Right hand small finger)
    }

    cursor(NO_CURSOR)   -- hide cursor while drawing

    -- erase old nails:
    set_color(TXT_COLOR, BACK_COLOR)
    for i = 1 to 10 do
	position(nailPos[i][1], LESSON_COLUMN - 1 + nailPos[i][2])
	puts(SCREEN, 32)
	position(nailPos[i][1] + 1, LESSON_COLUMN - 1 + nailPos[i][2])
	puts(SCREEN, 32)
    end for

    -- show current finger nail in colors:
    set_color(TEXT_NAIL_COLOR, BACK_NAIL_COLOR)
    position(nailPos[finger][1], LESSON_COLUMN - 1 + nailPos[finger][2])
    if key = ENTER then
	puts(SCREEN, 32)
    else
	puts(SCREEN, key)
    end if
    position(nailPos[finger][1] + 1, LESSON_COLUMN - 1 + nailPos[finger][2])
    puts(SCREEN, 32)
end procedure



-- PURPOSE: show the keyboard key for the current key in lesson.
-- USING:   set_color
procedure draw_key(integer key, integer finger)
    integer keyPos

    cursor(NO_CURSOR)   -- hide cursor while drawing

    -- erase old key:
    set_color(TXT_COLOR, BACK_COLOR)
    for line = 6 to 18 do
	position(line, 6)
	puts(SCREEN, LESSON_SCREEN[line])
    end for

    -- show key in colors:
    set_color(TEXT_NAIL_COLOR, BACK_NAIL_COLOR)

    -- print Space key:
    if key = 32 then
	position(18, LESSON_COLUMN + 31)
	puts(SCREEN, " Space ")

    -- print Enter key:
    elsif key = 13 then
	position(13, LESSON_COLUMN + 56)
	puts(SCREEN, " Enter ")

    -- print any other key:
    else
	-- find the key position in constant LESSON_SCREEN:
	for line = 6 to 16 do
	    keyPos = match(" " & key & " ", LESSON_SCREEN[line])

	    -- found the key position (row and column):
	    if keyPos then
		-- print the key on the keyboard:
		position(line, LESSON_COLUMN + keyPos)
		puts(SCREEN, 32 & key & 32)

		-- print Shift key of oposit hand if needed:
		if remainder(line, 3) = 0 then
		    -- print left or right Shift?
		    if finger > 5 then      -- print left Shift
			position(16, LESSON_COLUMN + 7)
		    else                    -- print right Shift
			position(16, LESSON_COLUMN + 55)
		    end if
		    puts(SCREEN, " Shift ")
		end if

		-- key is printed so exit:
		exit
	    end if
	end for
    end if
end procedure



-- PURPOSE: this is the main loop that runs the whole program.
-- USING:   draw_help_screen, draw_lesson_screen, choose_lesson,
--          create_lesson_string, draw_nail, draw_key, set_color
procedure main_input_loop()
    sequence lessonString
    integer lessonNumber, inKey, missed, exitLesson, resetStatus, finger
    atom startTime


    inKey = F5  -- start the program by letting user choose a lesson number.


    -- enter the main loop:
    while TRUE do

	-- display help screen:
	if inKey = F1 then
	    draw_help_screen()
	    inKey = wait_key()      -- press any key to exit help screen...
	    draw_lesson_screen(lessonNumber)    -- return to lesson screen

	-- choose another lesson:
	elsif inKey = F5 then
	    lessonNumber = choose_lesson()
	    draw_lesson_screen(lessonNumber)

	-- quit the program:
	elsif inKey = F10 then
	    if graphics_mode(-1) then end if    -- reset video
	    exit                                -- exit main loop (quit).
	end if


	-- get new string for current lesson:
	lessonString = create_lesson_string(lessonNumber)

	-- print the new string in the edit box (exclude the last ENTER key):
	set_color(TXT_COLOR, BACK_COLOR)
	position(3, LESSON_COLUMN + 2)
	puts(SCREEN, lessonString[1][1..STRING_LENGTH])

	-- initialize flags before starting a new lesson string:
	exitLesson = FALSE      -- reset exit flag
	resetStatus = TRUE      -- set flag


	-- let user type the keys for this lesson in the edit box:
	for i = 1 to STRING_LENGTH + 1 do  -- (+1 is for the last ENTER key).

	    -- draw nail and key on lesson screen for the current key:
	    finger = lessonString[2][i] - 48    -- (convert '1' to 1...)
	    if finger = 0 then                  -- fix index 0 to 10.
		finger = 10
	    end if
	    draw_nail(lessonString[1][i], finger)
	    draw_key(lessonString[1][i], finger)


	    -- enter the edit-one-key loop:
	    while TRUE do
		-- print 'Missed' and 'Timer' status:
		cursor(NO_CURSOR)
		set_color(TXT_COLOR, BACK_COLOR)
		position(1, LESSON_COLUMN + 30) -- ('Missed' status position).
		if resetStatus then
		    -- erase old 'Missed' and 'Timer':
		    puts(SCREEN, repeat(32, 42))
		else
		    -- first print 'Missed' status,
		    puts(SCREEN, sprintf("Missed: %d", missed))
		    -- then print 'Timer' status:
		    position(1, LESSON_COLUMN + 57)
		    puts(SCREEN, sprintf("Timer: %05d", time() - startTime))
		end if

		-- show the cursor in the edit box:
		cursor(UNDERLINE_CURSOR)
		position(3, LESSON_COLUMN + 1 + i)

		-- wait for user to press a key:
		inKey = wait_key()

		-- reset 'Missed' and 'Timer' status after user press the
		-- first key in edit box:
		if resetStatus then
		    missed = 0              -- reset 'Missed' status
		    startTime = time()      -- reset 'Timer' status
		    resetStatus = FALSE     -- reset flag after first input.
		end if

		-- user exit the lesson with any function key:
		if inKey = F1 or inKey = F5 or inKey = F10 then
		    exitLesson = TRUE          -- set exit flag
		    exit

		-- user typed the correct key, so print it in reverse colors
		-- and continue to the next key in string:
		elsif inKey = lessonString[1][i] then
		    set_color(BACK_COLOR, TXT_COLOR)
		    puts(SCREEN, lessonString[1][i])
		    exit

		-- user typed the wrong key, so add 1 to "Missed":
		else
		    missed += 1
		end if
	    end while

	    -- exit current lesson with function keys:
	    if exitLesson then
		exit
	    end if
	end for
    end while
end procedure



--bookmark: Start The Program;

    -- reset video to standard text mode (25 x 80):
    if graphics_mode(-1) then end if

    -- call the main loop to run the program:
    main_input_loop()

-- End of program.
